import os, sys

sys.path.insert(0, os.getcwd())

from unittest import TestCase

import pytest
import networkx as nx
import numpy as np
import random
import math

import cosasi


class TestShortFatTree(TestCase):
    def setUp(self):
        self.G = nx.random_tree(n=500, seed=0)
        contagion = cosasi.StaticNetworkContagion(
            G=self.G, model="si", infection_rate=0.1, number_infected=1
        )
        contagion.forward(50)
        self.t = 25
        self.I = contagion.get_infected_subgraph(self.t)
        return None

    def test_weighted_boundary_node_degree(self):
        # basic type check
        wbnd = cosasi.source_inference.single_source.weighted_boundary_node_degree(
            self.I, self.G, random.choice(list(self.I.nodes()))
        )
        assert isinstance(wbnd, (int, float))
        # double-check worked example
        G = nx.Graph()
        G.add_edges_from(
            [
                (1, 2),
                (2, 5),
                (2, 6),
                (2, 7),
                (1, 4),
                (4, 8),
                (4, 9),
                (4, 10),
                (1, 3),
                (3, 11),
            ]
        )
        I = G.subgraph([1, 2, 3, 4, 5])
        wbnd_1 = cosasi.source_inference.single_source.weighted_boundary_node_degree(
            I, G, 1, abs(math.log(0.5))
        )
        assert wbnd_1 == 0
        (
            wbnd_2,
            v_boundary,
        ) = cosasi.source_inference.single_source.weighted_boundary_node_degree(
            I, G, 2, abs(math.log(0.5)), True
        )
        assert sorted(v_boundary) == [3, 4]
        assert wbnd_2 > wbnd_1

    def test_short_fat_tree(self):
        result = cosasi.source_inference.single_source.short_fat_tree(self.I, self.G)
        assert isinstance(
            result, cosasi.source_inference.source_results.SingleSourceResult
        )
        noninf_vals = [i for i in result.data["scores"].values() if i != -np.inf]
        assert len(noninf_vals) <= len(
            self.I
        )  # score vals are wbnd, these are checked in test_weighted_boundary_node_degree



import os, sys

sys.path.insert(0, os.getcwd())

from unittest import TestCase

import pytest
import networkx as nx
import numpy as np
import random

import cosasi


class TestRumorCentrality(TestCase):
    def setUp(self):
        self.G = nx.random_tree(n=500, seed=0)
        contagion = cosasi.StaticNetworkContagion(
            G=self.G, model="si", infection_rate=0.1, number_infected=1
        )
        contagion.forward(50)
        self.t = 25
        self.I = contagion.get_infected_subgraph(self.t)
        return None

    def test_rumor_centrality_root(self):
        for _ in range(5):
            v = random.choice(list(self.I.nodes()))
            result_v = cosasi.single_source.rumor_centrality_root(self.I, v, False)
            assert isinstance(result_v, (int, float)) and result_v > 0
            result_dict = cosasi.single_source.rumor_centrality_root(self.I, v, True)
            for u in result_dict.keys():
                assert u in self.I.nodes()
                assert result_dict[u] > 0
        return None

    def test_rumor_centrality(self):
        with pytest.raises(ValueError):
            cosasi.single_source.rumor_centrality(self.I, self.G, "BAD INPUT")
        result = cosasi.single_source.rumor_centrality(
            self.I, self.G, None, False, False
        )
        assert isinstance(
            result, cosasi.source_inference.source_results.SingleSourceResult
        )
        result_data = result.data["scores"]
        assert isinstance(result_data, dict)
        for i in result_data.keys():
            assert i in self.G.nodes()
            assert isinstance(result_data[i], (float, int)) and result_data[i] > 0


    def test_rumor_centrality_root_example(self):
        """Verifies worked example from Section III.A of [1]_.

        References
        ----------
        .. [1] S., Devavrat and T. Zaman,
            "Rumors in a network: Who's the culprit?."
            IEEE Transactions on Information Theory, 2011
            https://devavrat.mit.edu/wp-content/uploads/2017/10/Rumors-in-a-network-whos-the-culprit.pdf
        """
        I = nx.Graph()
        I.add_edges_from([(1, 2), (1, 3), (2, 4), (2, 5)])
        assert (
            cosasi.single_source.rumor_centrality_root(I, 1, False)
            == cosasi.single_source.rumor_centrality_root(I, 1, True)[1]
            == 8
        )
        # assert cosasi.single_source.rumor_centrality_root(I, 1, True)[1] == 8



import os, sys

sys.path.insert(0, os.getcwd())

from unittest import TestCase

import pytest
import networkx as nx
import numpy as np
import random

import cosasi


class TestEarliestInfectionFirst(TestCase):
    def setUp(self):
        self.G = nx.fast_gnp_random_graph(100, 0.25)
        contagion = cosasi.StaticNetworkContagion(
            G=self.G, model="si", infection_rate=0.01, number_infected=1
        )
        contagion.forward(50)
        self.t = 20
        self.I = contagion.get_infected_subgraph(self.t)
        self.observers = contagion.get_observers(10)
        return None

    def test_earliest_infection_first_disconnected(self):
        H = nx.disjoint_union(self.G, self.G)
        with pytest.raises(ValueError):
            cosasi.single_source.earliest_infection_first(
                I=self.G, G=H, observer_dict=self.observers
            )

    def test_earliest_infection_first(self):
        result = cosasi.source_inference.single_source.earliest_infection_first(
            I=self.I, G=self.G, observer_dict=self.observers
        )
        assert isinstance(
            result, cosasi.source_inference.source_results.SingleSourceResult
        )
        result_data = result.data["scores"]
        assert isinstance(result_data, dict)
        for i in result_data.keys():
            assert i in self.G.nodes()
            assert isinstance(result_data[i], (float, int)) and result_data[i] > 0



import os, sys

sys.path.insert(0, os.getcwd())

from unittest import TestCase

import pytest
import networkx as nx
import numpy as np

import cosasi


class TestEstimators(TestCase):
    def setUp(self):
        self.G = self.G = nx.gnp_random_graph(50, 0.2)
        contagion = cosasi.StaticNetworkContagion(
            G=self.G, model="si", infection_rate=0.1, number_infected=1
        )
        contagion.forward(30)
        self.t = 25
        self.I = contagion.get_infected_subgraph(self.t)
        return None

    def test_source_subgraphs(self):
        for i in range(1, 10):
            subgraphs = cosasi.utils.estimators.source_subgraphs(
                self.G, number_sources=i
            )
            assert len(subgraphs) == i

    def test_number_sources(self):
        for method in ["eigengap", "netsleuth", "chatter"]:
            # check that number of sources matches when provided
            for number_sources in range(1, 5):
                n, subgraphs = cosasi.utils.estimators.number_sources(
                    I=self.I,
                    number_sources=number_sources,
                    return_source_subgraphs=True,
                    number_sources_method=method,
                    G=self.G,
                )
                for g in subgraphs:
                    assert type(g) == nx.Graph
                assert n == number_sources
            # estimating numbers as expected
            n, subgraphs = cosasi.utils.estimators.number_sources(
                I=self.I,
                number_sources=None,
                return_source_subgraphs=True,
                number_sources_method=method,
                G=self.G,
            )
            m = cosasi.utils.estimators.number_sources(
                I=self.I,
                number_sources=None,
                return_source_subgraphs=False,
                number_sources_method=method,
                G=self.G,
            )
            assert n == m and (isinstance(n, np.int64) or isinstance(n, int))
            # just return number_sources back
            n = 2
            m = cosasi.utils.estimators.number_sources(
                I=self.I,
                number_sources=n,
                return_source_subgraphs=False,
                number_sources_method=method,
                G=self.G,
            )
            assert n == m
        # check error-handling
        with pytest.raises(NotImplementedError):
            assert cosasi.utils.estimators.number_sources(
                self.I,
                number_sources=None,
                return_source_subgraphs=False,
                number_sources_method="BAD INPUT",
            )
        with pytest.raises(ValueError):
            # Need `G` for NETSLEUTH method
            assert cosasi.utils.estimators.number_sources(
                G=None,
                I=self.I,
                number_sources=None,
                return_source_subgraphs=False,
                number_sources_method="netsleuth",
            )
        with pytest.raises(ValueError):
            # Need `G` for chatter method
            assert cosasi.utils.estimators.number_sources(
                G=None,
                I=self.I,
                number_sources=None,
                return_source_subgraphs=False,
                number_sources_method="chatter",
            )
        with pytest.raises(ValueError):
            # Need `G` for chatter method
            assert cosasi.utils.estimators.number_sources(
                G=self.G,
                I=self.I,
                number_sources="BAD INPUT",
                return_source_subgraphs=False,
                number_sources_method="chatter",
            )

    def test_eigengap(self):
        assert isinstance(cosasi.utils.estimators.eigengap(self.G), np.int64)
        # two disjoint complete graphs should have a spectral gap of 2
        K = nx.complete_graph(10)
        H = nx.disjoint_union(K, K)
        assert cosasi.utils.estimators.eigengap(H) == 2

    def test_bits_encode_integer(self):
        last = 0
        for i in range(-10, 10):
            if i < 1:
                with pytest.raises(ValueError):
                    cosasi.utils.estimators.bits_encode_integer(i)
            else:
                # number of bits should increase w/ integer
                bits = cosasi.utils.estimators.bits_encode_integer(i)
                assert bits > last
                last = bits
        assert cosasi.utils.estimators.bits_encode_integer(1) == pytest.approx(1.05259)

    def test_bits_encode_seed(self):
        seed = [1, 2, 3]
        assert cosasi.utils.estimators.bits_encode_seed(
            seed, self.G
        ) > cosasi.utils.estimators.bits_encode_integer(len(seed))

    def test_bits_encode_ripple(self):
        bits_ripple_1 = cosasi.utils.estimators.bits_encode_ripple(
            list(range(1)), self.G
        )
        bits_ripple_2 = cosasi.utils.estimators.bits_encode_ripple(
            list(range(5)), self.G
        )
        assert 0 < min(bits_ripple_1, bits_ripple_2)

    def test_description_length(self):
        seed = [1, 2, 3]
        assert (
            min(
                cosasi.utils.estimators.description_length(seed, self.G),
                cosasi.utils.estimators.description_length(seed, self.G),
            )
            > 0
        )

    def test_chatter_frequency(self):
        message_frequency = cosasi.utils.estimators.chatter_frequency(self.G, 5)
        assert message_frequency.size == len(self.G) * len(self.G)
        assert np.min(message_frequency) >= 0

    def test_chatter_distance(self):
        for t in [1, 5, None]:
            for G in [self.G, nx.disjoint_union(self.G, self.G)]:
                # distances are non-negative
                assert (
                    np.min(
                        cosasi.utils.estimators.chatter_distance(
                            G=G, t=t, normalized=False
                        )
                    )
                    >= 0
                )
                dist = cosasi.utils.estimators.chatter_distance(
                    G=G, t=t, normalized=True
                )
                if (
                    dist[0][1] != dist[0][1]
                    or np.max(dist) != np.max(dist)
                    or np.min(dist) != np.min(dist)
                ):
                    pass
                else:
                    assert dist[0][1] == cosasi.utils.estimators.chatter_distance(
                        G=G, t=t, u=0, v=1, normalized=True
                    )
                    # check normalization
                    assert 1 >= np.max(dist) >= np.min(dist) >= 0
                    # array ops work right
                    assert len(
                        cosasi.utils.estimators.chatter_distance(G=G, t=t, u=0)
                    ) == len(G)
                    assert len(
                        cosasi.utils.estimators.chatter_distance(self.G, 5, v=0)
                    ) == len(G)
                    # should be symmetric
                    for i in range(len(G)):
                        for j in range(len(G)):
                            assert cosasi.utils.estimators.chatter_distance(
                                G=G, t=t, u=i, v=j, normalized=False
                            ) == cosasi.utils.estimators.chatter_distance(
                                G=G, t=t, u=j, v=i, normalized=False
                            )

    def test_chatter(self):
        result = cosasi.utils.estimators.chatter(self.I, self.G)
        assert isinstance(result, int)
        assert result > 0
        assert result <= len(self.I)



import os, sys
import collections

sys.path.insert(0, os.getcwd())

import pytest
from unittest import TestCase
import networkx as nx
import numpy as np

import cosasi


class Test_StaticNetworkContagion(TestCase):
    def setUp(self):
        self.number_infected_init = 10
        self.sim_steps = 10
        self.G = nx.fast_gnp_random_graph(200, 0.25)
        self.contagion = cosasi.StaticNetworkContagion(
            G=self.G,
            model="si",
            infection_rate=0.1,
            number_infected=self.number_infected_init,
        )
        self.contagion.forward(self.sim_steps)
        return None

    def test_argument_exceptions(self):
        with pytest.raises(ValueError):
            # G must be a NetworkX graph
            cosasi.StaticNetworkContagion(G="BAD INPUT", model="si", infection_rate=0.1)
        with pytest.raises(NotImplementedError):
            # model must be "sir", "si", or "sis"
            cosasi.StaticNetworkContagion(
                G=self.G, model="BAD INPUT", infection_rate=0.1
            )
        with pytest.raises(ValueError):
            # infection_rate must be between 0 and 1
            cosasi.StaticNetworkContagion(G=self.G, model="si", infection_rate=10)
        with pytest.raises(ValueError):
            # recovery_rate must be between 0 and 1
            cosasi.StaticNetworkContagion(
                G=self.G, model="si", infection_rate=0.1, recovery_rate=10
            )
        with pytest.raises(ValueError):
            # can only provide one of fraction_infected, number_infected
            cosasi.StaticNetworkContagion(
                G=self.G,
                model="si",
                infection_rate=0.1,
                fraction_infected=0.1,
                number_infected=self.number_infected_init,
            )
        for m in ["sir", "sis"]:
            with pytest.raises(ValueError):
                # requires recovery rate
                cosasi.StaticNetworkContagion(G=self.G, model=m, infection_rate=0.1)
            cosasi.StaticNetworkContagion(
                G=self.G, model=m, infection_rate=0.1, recovery_rate=0.05
            )
            assert True

    def test_fraction_infected(self):
        contagion = cosasi.StaticNetworkContagion(
            G=self.G,
            model="si",
            infection_rate=0.1,
            fraction_infected=self.number_infected_init / len(self.G),
        )
        contagion.forward(5)
        assert self.number_infected_init == len(contagion.get_source())

    def test_get_infected_indices(self):
        assert len(self.contagion.get_infected_indices()) == self.number_infected_init
        temp_contagion = cosasi.StaticNetworkContagion(
            G=self.G,
            model="si",
            infection_rate=0.1,
            fraction_infected=None,
            number_infected=None,
        )
        temp_contagion.forward()
        assert len(temp_contagion.get_infected_indices()) == 1
        return None

    def test_forward(self):
        assert len(self.contagion.history) == self.sim_steps
        return None

    def test_reset_sim(self):
        self.contagion.reset_sim()
        assert len(self.contagion.history) == 0
        self.contagion.forward(self.sim_steps)
        return None

    def test_get_infected_subgraph(self):
        sg = self.contagion.get_infected_subgraph(step=self.sim_steps - 1)
        assert isinstance(sg, nx.Graph)
        assert len(sg) == len(
            self.contagion.get_infected_indices(step=self.sim_steps - 1)
        )
        assert set(sg.nodes) == set(
            self.contagion.get_infected_indices(step=self.sim_steps - 1)
        )  # sets are unordered
        return None

    def test_get_observers(self):
        num_observers = 5
        self.contagion.forward(steps=100)
        observers = self.contagion.get_observers(observers=num_observers)
        assert len(observers) == num_observers  # check size
        for i in observers.keys():
            assert i in self.G.nodes
            # check types
            assert isinstance(observers[i], (int, float, type(None)))
            if isinstance(observers[i], float):
                observers[i] == np.inf
        return None

    def test_get_source(self):
        source_verts = self.contagion.get_source()
        assert isinstance(source_verts, list)
        assert len(source_verts) <= len(self.G)
        source_graph = self.contagion.get_source(return_subgraph=True)
        assert isinstance(source_graph, nx.Graph)
        assert set(source_graph.nodes) == set(source_verts)
        return None

    def test_get_frontier(self):
        s = 15
        G = nx.fast_gnp_random_graph(100, 0.25)
        contagion = cosasi.StaticNetworkContagion(
            G=G, model="si", infection_rate=0.01, number_infected=3
        )
        contagion.forward(500)
        I = contagion.get_infected_subgraph(step=s)
        frontier = contagion.get_frontier(step=s)
        assert isinstance(frontier, (list, set))  # basic type checking
        assert all(
            [v in I for v in frontier]
        )  # frontier is a subset of infection subgraph
        beyond_frontier = []
        for i in I:
            if i not in frontier:
                # every node not in the frontier has all neighbors in I
                assert all([v in I for v in G.neighbors(v)])
        # frontier at step 0 should just be the initially-infected indices
        frontier_0 = contagion.get_frontier(step=0)
        assert collections.Counter(frontier_0) == collections.Counter(
            contagion.get_source()
        )
        # when the graph is saturated/maximally-infected, frontier should be empty
        largest_cc_size = len(max(nx.connected_components(G), key=len))
        while len(contagion.get_infected_indices(step=s)) < largest_cc_size:
            s += 1
        assert contagion.get_frontier(step=s) == set()



import os, sys

sys.path.insert(0, os.getcwd())

from unittest import TestCase

import pytest
import networkx as nx
import numpy as np
import random

import cosasi


class TestLISN(TestCase):
    def setUp(self):
        self.G = nx.random_tree(n=500, seed=0)
        contagion = cosasi.StaticNetworkContagion(
            G=self.G, model="si", infection_rate=0.1, number_infected=1
        )
        contagion.forward(50)
        self.t = 25
        self.I = contagion.get_infected_subgraph(self.t)
        return None

    def test_distance_prob(self):
        # for constant distance, probability should weakly increase with time
        last_prob = -np.inf
        n = random.randint(1, 10)
        for t in range(1, 10):
            prob = cosasi.source_inference.single_source.distance_prob(t, n, 0.05)
            assert prob >= last_prob
            last_prob = prob
        # for constant time, probability should weakly decrease with distance
        last_prob = np.inf
        t = random.randint(1, 10)
        for n in range(1, 10):
            prob = cosasi.source_inference.single_source.distance_prob(t, n, 0.05)
            assert prob <= last_prob
            last_prob = prob

    def test_lisn(self):
        result = cosasi.source_inference.single_source.lisn(self.I, self.G, self.t)
        # type check
        assert isinstance(
            result, cosasi.source_inference.source_results.SingleSourceResult
        )
        # -inf only for nodes outside infection subgraph
        vals = list(result.data["scores"].values())
        noninf_vals = [i for i in result.data["scores"].values() if i != -np.inf]
        assert len(noninf_vals) == len(self.I) and len(vals) == len(self.G)
        # scores are log probabilities
        assert -np.inf <= max(noninf_vals) <= max(noninf_vals) <= 0
